from ipykernel.zmqshell import ZMQInteractiveShell
from .output import OutputStreamsSetter, OutputText, OutputTextStyled

import sys


class IpythonWatcher(object):
    def __init__(self, ip: ZMQInteractiveShell):
        self.shell = ip
        self._save_prev_outputs()

    def _save_prev_outputs(self):
        self.stdout = sys.stdout
        self.stderr = sys.stderr

    def _activate_prev_outputs(self):
        sys.stdout = self.stdout
        sys.stderr = self.stderr

    def pre_execute(self):
        self._save_prev_outputs()
        stream = OutputStreamsSetter(stdout=OutputText(kind='stdout'), stderr=OutputText(kind='stderr'))
        stream.activate()

    def pre_run_cell(self, info):
        ...

    def post_execute(self):
        self._activate_prev_outputs()

    def post_run_cell(self, result):
        ...


def register_hooks(watcher: IpythonWatcher, ip: ZMQInteractiveShell):
    ip.events.register("pre_execute", watcher.pre_execute)
    ip.events.register("pre_run_cell", watcher.pre_run_cell)
    ip.events.register("post_execute", watcher.post_execute)
    ip.events.register("post_run_cell", watcher.post_run_cell)


def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    from .magic import IgogoMagic
    ipython.register_magics(IgogoMagic)



import io
class IgogoInvalidContext(Exception):
    def __init__(self):
        from .core import _log_error
        file = io.StringIO()
        _log_error('Igogo command invoked from invalid context', file=file)
        super().__init__(file.getvalue())
        file.close()


class IgogoAdditionalOutputsExhausted(Exception):
    def __init__(self):
        from .core import _log_error
        file = io.StringIO()
        _log_error('Igogo invoked display function, but cell has not enough display capabilities', file=file)
        super().__init__(file.getvalue())
        file.close()



import random
import string
import io
import sys
from IPython import display
from IPython.core.interactiveshell import InteractiveShell


def is_lab_notebook():
    import re
    import psutil

    return any(re.search('jupyter-lab', x)
               for x in psutil.Process().parent().cmdline())


class OutputBase:
    def __init__(self, display_id=None):
        if display_id is None:
            self.display_id = ''.join(random.choices(string.ascii_uppercase + string.digits, k=100))
        else:
            self.display_id = display_id

        self.handle = display.display(display_id=self.display_id)
        self.metadata = dict()
        self.dic_kind = {
            'text': 'text/plain',
            'markdown': 'text/markdown',
            'html': 'text/html',
            'stderr': 'application/vnd.jupyter.stderr',
            'stdout': 'application/vnd.jupyter.stdout'
        }
        self.objs = []
        if not is_lab_notebook():
            self.display()

    def display(self, update=False):
        display.display(
            *self.objs,
            display_id=self.display_id,
            update=update,
            raw=True,
            metadata=self.metadata,
            transient={
                'display_id': self.display_id
            }
        )

    def _build_obj(self, content, kind):
        mime_type = self.dic_kind.get(kind, kind)
        print({mime_type: content}, file=open('log.txt', 'w'))
        return {mime_type: content}

    def _update(self):
        self.display(update=True)

    def clear(self):
        self.text = ''
        self.objs = []
        self.metadata = dict()
        self._update()


class OutputText(OutputBase, io.IOBase):

    def __init__(self, kind, display_id=None):
        super().__init__(display_id)
        self.text = ''
        self.kind = kind

    def add_text(self, content):
        self.text += content
        self.objs = [self._build_obj(self.text, self.kind)]
        self._update()

    def flush(self): ...

    def write(self, data):
        self.add_text(data)


class OutputObject(OutputBase):
    def add_object(self, obj, include=None, exclude=None):
        fmt = InteractiveShell.instance().display_formatter.format
        format_dict, md_dict = fmt(obj, include=include, exclude=exclude)
        if not format_dict:
            return
        self.objs = [format_dict]
        self.metadata = md_dict
        self._update()


class OutputTextStyled(OutputBase, io.IOBase):
    def __init__(self, style_start='', style_end='', kind='html'):
        super().__init__()
        self.text = ''
        self.style_end = style_end
        self.style_start = style_start
        self.kind = kind

    def _build_styled(self, content):
        type = self.dic_kind.get(self.kind)
        content = self.style_start + content + self.style_end
        return {type: content}

    def add_text(self, content):
        self.text += content
        self.objs = [self._build_styled(self.text)]
        self._update()

    def write(self, data):
        self.add_text(data)

    def flush(self): ...


class OutputStreamsSetter:
    def __init__(self, stdout: OutputText, stderr: OutputText):
        self.stdout = stdout
        self.stderr = stderr

    def activate(self):
        sys.stdout = self.stdout
        sys.stderr = self.stderr



from typing import List
import asyncio
import contextvars

from .output import OutputStreamsSetter, OutputObject
from .exceptions import IgogoInvalidContext


class IgogoContext(object):
    out_stream: OutputStreamsSetter
    task: asyncio.Task
    additional_outputs: List[OutputObject]

    def __init__(self, task: asyncio.Task, out_stream: OutputStreamsSetter, additional_outputs: List[OutputObject]):
        self.out_stream = out_stream
        self.task = task
        self.additional_outputs = additional_outputs


_context: contextvars.ContextVar[IgogoContext | None] = contextvars.ContextVar("igogo_context", default=None)


def get_context_or_none() -> IgogoContext:
    return _context.get()


def get_context_or_fail() -> IgogoContext:
    value = get_context_or_none()
    if value is None:
        raise IgogoInvalidContext()
    return value


def set_context(context: IgogoContext):
    _context.set(context)



import IPython
from IPython.core.magic import (Magics, magics_class, cell_magic)

@magics_class
class IgogoMagic(Magics):
    @cell_magic
    def igogo(self, line, cell):
        ip = IPython.get_ipython()
        args: dict = eval(f"dict({line})")
        args.setdefault('update_globals', False)
        update_globals = args['update_globals']
        args.pop('update_globals')
        prefix = "def __igogo_magic_wrapper():\n" \
                 "    import igogo\n" \
                 f"    @igogo.job(**dict({args}))\n" \
                 "    async def execute():\n"\
                 "        global " + ', '.join(list(IPython.get_ipython().user_ns.keys())) + '\n'
        cell = prefix + '\n'.join(['        ' + line for line in cell.split('\n')])
        cell += "\n" + \
               ("        globals().update(locals())\n" if update_globals else "") + \
                "    return execute()\n" \
                "__igogo_magic_wrapper()"
        ip.run_cell(cell, silent=True)




import asyncio
import functools
import inspect
import sys
from typing import Dict, List

import IPython
import greenback

from .context import IgogoContext, get_context_or_fail, set_context
from .output import OutputText, OutputStreamsSetter, OutputObject, OutputTextStyled
from .exceptions import IgogoInvalidContext, IgogoAdditionalOutputsExhausted

_igogo_run_loop = asyncio.get_running_loop()
_all_tasks: Dict[int, List[asyncio.Task]] = dict()
_igogo_count = 0


def _get_currently_running_cells_info():
    global _all_tasks
    keys = map(str, _all_tasks.keys())
    keys = '], ['.join(list(keys))
    if len(keys) > 0:
        keys = '[' + keys + ']'
    return keys


def _log_error(*argc, **kwargs):
    if not 'file' in kwargs:
        kwargs['file'] = sys.stderr
    print('[ IGOGO ]', *argc, **kwargs)
    running_s = _get_currently_running_cells_info()
    if len(running_s) == 0:
        running_s = '<none>'
    print(f'[ IGOGO ] Currently running IGOGO cells: {running_s}', file=kwargs['file'])


def stop():
    value = get_context_or_fail()
    value.task.cancel()


def get_running_igogo_cells():
    global _all_tasks
    _update_all_tasks()
    return list(_all_tasks.keys())


def sleep(delay, result=None):
    if not greenback.has_portal():
        raise IgogoInvalidContext()
    greenback.await_(asyncio.sleep(delay, result))


def display(object):
    value = get_context_or_fail()
    if len(value.additional_outputs) == 0:
        raise IgogoAdditionalOutputsExhausted()
    out = value.additional_outputs.pop()
    out.add_object(object)
    value.additional_outputs.insert(0, out)


def clear_output(including_text=True):
    value = get_context_or_fail()
    if including_text:
        value.out_stream.stdout.clear()
    for out in value.additional_outputs:
        out.clear()


def _update_all_tasks():
    global _all_tasks

    def filter_rule(task: asyncio.Task):
        return not task.done()

    for key in _all_tasks:
        _all_tasks[key] = list(filter(filter_rule, _all_tasks[key]))
    _all_tasks = {k: v for k, v in _all_tasks.items() if len(v) > 0}


def get_pending_tasks():
    return list(filter(lambda x: 'igogo' in x.get_name(), asyncio.all_tasks(loop=_igogo_run_loop)))


def stop_all():
    for task in get_pending_tasks():
        task.cancel()


def stop_latest():
    global _all_tasks
    _update_all_tasks()
    keys = list(_all_tasks.keys())
    if len(keys) == 0:
        _log_error("No running tasks")
        return
    latest_key = max(keys)
    task = _all_tasks[latest_key].pop()
    task.cancel()


def stop_by_cell_id(cell_id):
    global _all_tasks
    _update_all_tasks()

    cell_id = int(cell_id)
    if not cell_id in _all_tasks:
        _log_error(f"There's no running tasks in cell [{cell_id}]")
        return
    for task in _all_tasks[cell_id]:
        task.cancel()


def job(original_function=None, kind='stdout', displays=10):
    global _igogo_count

    def _decorate(function):
        @functools.wraps(function)
        def wrapped_function(*args, **kwargs):
            global _igogo_count, _all_tasks
            output_stream = OutputStreamsSetter(stdout=OutputText(kind=kind), stderr=OutputText(kind='stderr'))
            additional_outputs = list(reversed([OutputObject() for _ in range(displays)]))

            async def func_context_setter():
                await greenback.ensure_portal()
                set_context(
                    IgogoContext(task, output_stream, additional_outputs)
                )
                output_stream.activate()
                if inspect.iscoroutinefunction(function):
                    return await function(*args, **kwargs)
                return function(*args, **kwargs)

            coro = func_context_setter()

            if not hasattr(wrapped_function, "tasks"):
                wrapped_function.tasks = []

            task = _igogo_run_loop.create_task(coro)
            wrapped_function.tasks.append(task)

            def done_callback(t):
                try:
                    exception = task.exception()
                    if exception is not None:
                        raise exception
                except asyncio.CancelledError:
                    pass

            task.add_done_callback(done_callback)
            task.set_name(f'igogo_{_igogo_count}')
            _igogo_count += 1

            ip = IPython.get_ipython()
            _all_tasks.setdefault(ip.execution_count, [])
            _all_tasks[ip.execution_count].append(task)

            return dict(
                task=task
            )

        def stop_all():
            if not hasattr(wrapped_function, "tasks"):
                wrapped_function.tasks = []
            for task in wrapped_function.tasks:
                task.cancel()

        from .yielder import Yielder
        wrapped_function.yielder = Yielder
        wrapped_function.stop_all = stop_all
        wrapped_function.stop = stop
        return wrapped_function

    if original_function:
        return _decorate(original_function)
    return _decorate



from .context import get_context_or_none, get_context_or_fail
from .exceptions import IgogoInvalidContext

import greenback


class Yielder:

    @classmethod
    def igogo_await(cls):
        if not greenback.has_portal() and get_context_or_none() is None:
            return
        elif not greenback.has_portal():
            raise IgogoInvalidContext()
        greenback.await_(cls())
        value = get_context_or_fail()
        value.out_stream.activate()

    def __await__(self):
        try:
            get_context_or_fail()
        except IgogoInvalidContext:
            return
        yield
        value = get_context_or_fail()
        value.out_stream.activate()



import IPython
from .yielder import Yielder as yielder_async
from .core import job, stop, sleep, display, clear_output, stop_all
from .core import stop_by_cell_id, stop_latest, get_running_igogo_cells
from .loaders import register_hooks as _register_hooks
from .loaders import IpythonWatcher
from .loaders import load_ipython_extension

yielder = yielder_async.igogo_await

__igogo_inited = False

if not __igogo_inited:
    __igogo_inited = True
    _ip = IPython.get_ipython()
    _watcher = IpythonWatcher(_ip)
    _register_hooks(_watcher, _ip)
